# Copyright dr. ir. R.J.F. Bijster, 2022

from enum import Enum
from enum import IntEnum
import random
import numpy as np
from scipy.optimize import fsolve

from SimulationConstants import *


class BicycleState(IntEnum):
    STOP = 0
    ACCELERATING = 1
    CRUISING = 2
    DECELERATING = 3
    FINISHED = 4

class Bicycle():
    def __init__(self, ridermass, bikemass, maxpower = 100, timestep = 1, speedlimiter = np.inf, acceleration = 1):
        self.position = 0                   # m, current position
        self.speed = 0                      # m/s, instantaneous speed
        self.power = 0                      # W, instantaneous power
        self.acceleration = acceleration    # m/s2, acceleration rate
        self.deceleration_comfort = 1.5     # m/s^2, Comfortable deceleration rate for normal bicycle, lower bound (CROW 2006)
        self.deceleration_emergency = 2.6   # m/s^2, Emergency deceleration rate for normal bicycle (CROW 2006)
        self.maxspeed = 0                   # m/s, Maximum speed that can be reached with the available power
        self.target = 0
        self.mass = ridermass + bikemass
        self.maxpower = maxpower
        self.state = BicycleState.STOP
        self.timestep = timestep
        self.speedlimiter = speedlimiter    # m/s, If the bicycle has a limiter, for example for a pedelec system

    def calculateTopSpeed(self, windspeed, inclination):

        c1 = 1/2*rho*ApCd                           # Air drag
        c2 = cr * self.mass * g                     # Rolling resistance
        c3 = self.mass * g * np.sin(inclination)    # inclination

        # Estimate the average speed of cyclist on a city bike, with no wind
        func = lambda v : self.maxpower - (c1*(v+windspeed)**2*v + c2*v + c3*v)

        v_initial_guess = 5.5
        v_cruise = fsolve(func, v_initial_guess)
        
        self.maxspeed = float(max(0, min(v_cruise, self.speedlimiter)))


    def calculateBrakingDistance(self, speed, deceleration):
        '''Calculate braking distance from current speed at given deceleration rate'''
        time_to_stop = speed/deceleration
        distance_to_stop = 1/2*deceleration*time_to_stop**2 
        # distance_to_stop = speed**2 / (2*deceleration)
        return distance_to_stop

    def calculateRequiredDeceleration(self, speed, distance, p = False):
        # distance = 1/2 * a * t^2 
        if speed > 0:
            time_to_stop = distance/(speed/2)
            deceleration = 2*distance / time_to_stop**2
            if p == True:
                print(f'Distance: {distance} m, Velocity: {speed} m/s, Time to stop: {time_to_stop} s, Decelerate at: {deceleration} m/s2')
        else: 
            deceleration = 0
        # deceleration = 2 * speed**2 / distance
        return deceleration

    def ride(self, trail):

        if self.state == BicycleState.FINISHED:
            return

        self.calculateTopSpeed(windspeed = trail.windspeed, inclination = trail.inclination)
        nextWP = trail.getNextWayPoint(location = self.position)
        distance_to_waypoint = nextWP.getLocation() - self.position
        comfortable_braking_distance = self.calculateBrakingDistance(self.maxspeed, self.deceleration_comfort)
        required_braking_rate        = max(self.calculateRequiredDeceleration(self.speed, distance_to_waypoint), 0.1)

        self.target = trail.getLength()

        if nextWP.type == WaypointType.ENDPOINT:
            # Continue pedaling, you're almost there.
            self.update(acceleration = self.acceleration)
        elif nextWP.type == WaypointType.SIGNAL:
            if nextWP.state == SignalState.ORANGE or nextWP.state == SignalState.RED:
                if (distance_to_waypoint <= comfortable_braking_distance) and (required_braking_rate <= self.deceleration_emergency):
                    self.update(acceleration = np.abs(required_braking_rate), brake = True)
                    return
            # Continue
            self.update(acceleration = self.acceleration)
                    
    def update(self, acceleration, brake = False):

        if self.state == BicycleState.FINISHED:
            # We are already there yet. Don't do anything.
            return

        if (self.state == BicycleState.STOP) and brake:
            # We are already standing still. Don't do anything.
            self.speed = 0
            self.updatePosition()
            return

        if (self.state == BicycleState.STOP or self.state == BicycleState.ACCELERATING or self.state == BicycleState.DECELERATING) and not brake:
            speed = self.speed + self.acceleration*self.timestep
            if speed >= self.maxspeed:
                self.state = BicycleState.CRUISING
                self.speed = self.maxspeed
            else:
                self.speed = speed
                self.state = BicycleState.ACCELERATING
            self.updatePosition()
            return
        
        if (self.state == BicycleState.ACCELERATING or self.state == BicycleState.CRUISING or self.state == BicycleState.DECELERATING) and brake:
            speed = self.speed - acceleration*self.timestep
            if speed <= 0:
                self.state = BicycleState.STOP
                self.speed = 0
            else:
                self.speed = speed
                self.state = BicycleState.DECELERATING
            self.updatePosition()
            return

        if (self.state == BicycleState.CRUISING) and not brake:
            self.updatePosition()
            return

    def updatePosition(self):
        position = self.position + self.speed*self.timestep
        if position >= self.target:
            self.state = BicycleState.FINISHED
            self.position = self.target 
            self.speed = 0
        else:
            self.position = position


    def reset(self):
        self.position = 0
        self.speed = 0
        self.power = 0


    def getStats(self):
        return self.position, self.speed, self.state
    
class SignalState(IntEnum):
    RED = 0
    GREEN = 1
    ORANGE = 2
    NOTDEFINED = 3

class WaypointType(IntEnum):
    ENDPOINT = 0
    SIGNAL = 1

class Waypoint():
    def __init__(self, location = 0):
        # Location
        self.location = location 
        self.type = WaypointType.ENDPOINT
        self.state = SignalState.NOTDEFINED

    def getLocation(self):
        return self.location

    def update(self, time):
        pass

class SignaledIntersection(Waypoint):

    def __init__(self, location = 0, cycleTime = 90, greenTime = 8, orangeTime = 3):
        super().__init__(location)
        self.type = WaypointType.SIGNAL

        self.cycleTime = cycleTime      # s, time for a full cycle at the intersection

        self.greenTime = greenTime      # s, time the signal is green
        self.orangeTime = orangeTime    # s, time the signal is orange
        self.redTime = self.cycleTime - self.greenTime - self.orangeTime

        # Initialize state
        self.state = SignalState.RED
        self.phaseTime = 0
        self.offset = 0
        
    def update(self, time):
        self.phaseTime = (time + self.offset) % self.cycleTime
        self.determineState()
    
    def determineState(self):
        '''Determines the state based on the phase'''
        
        greenStartsAt = self.redTime
        orangeStartsAt = self.redTime + self.greenTime

        if (self.phaseTime >= greenStartsAt) and (self.phaseTime < orangeStartsAt):
            self.state = SignalState.GREEN
        elif self.phaseTime >= orangeStartsAt:
            self.state = SignalState.ORANGE
        else:
            self.state = SignalState.RED

    def initializeAtRandomPhase(self):
        '''Pick a random moment (normalized) in the sequence to start'''
        # Standard sequence: RED, GREEN, ORANGE (repeat)
        # Generate a random starting point in the phase
        phase = random.random()
        self.phaseTime = phase * self.cycleTime
        self.offset = self.phaseTime

        self.determineState()

class Trail():
    def __init__(self, length_km, numlegs = 1, windspeed = 0, inclination = 0, timestep = 1):
        self.length_km = length_km
        self.length = length_km * 1000
        self.numlegs = numlegs
        self.climb = 0
        self.windspeed = windspeed
        self.inclination = inclination
        self.waypoints = []
        self.time = 0
        self.timestep = timestep

        # Generate
        self.generateTrail()


    def getLength(self):
        return self.length

    def getInclination(self):
        return self.inclination

    def generateTrail(self):
        '''Generate the trail in n legs, with n-1 signals in between.'''

        # Generate the locations where the signals are going to be. 
        locs = self.randompoints(self.numlegs - 1, self.length)
        
        # Add the signals to the array of signals
        for loci in locs:
            signal = SignaledIntersection(location = loci)
            signal.initializeAtRandomPhase()
            self.waypoints.append(signal)
        # Add the end point and turn the entire list into waypoints
        endpoint = Waypoint(location = self.length)
        self.waypoints.append(endpoint)
        
    def randompoints(self, num, L = 1):
        '''Generate num points on [0, 1). Sort them from small to large'''
        points = []
        for i in range(num):
            points.append(random.random()*L)
        points.sort()
        return points
    
    def getNextWayPoint(self, location):
        nextWP = 0
        for wp in reversed(self.waypoints):
            wplocation = wp.getLocation()
            
            if location > wplocation:
                break
            else: 
                nextWP = wp
                

        return nextWP

    def getWayPoints(self):
        return self.waypoints

    def getWayPointsCoordsAndState(self):
        x = []
        states = []
        for wp in self.waypoints:
            x.append(wp.getLocation())
            states.append(wp.state)

        return x, states

    def update(self):
        self.time = self.time + self.timestep
        for wp in self.waypoints:
            wp.update(self.time)